import argparse

from petisco import __version__


def has_args(args):
    is_active = False
    for arg in vars(args):
        is_active = is_active or getattr(args, arg)
    return is_active


def main():
    parser = argparse.ArgumentParser(
        prog="petisco 🍪",
        description="petisco is a framework for helping Python developers to build clean Applications",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "-v", "--version", action="store_true", help="show petisco version number."
    )

    args = parser.parse_args()

    if not has_args(args):
        parser.print_help()
    else:
        if args.version:
            print(f"petisco 🍪 => {__version__}")
            return



import inspect
from typing import Callable, Dict, Any

from dataclasses import dataclass
from dotmap import DotMap

from petisco.events.interface_event_manager import IEventManager
from petisco.events.service_deployed import ServiceDeployed
from petisco.frameworks.interface_application import IApplication
from petisco.logger.interface_logger import INFO, ILogger
from petisco.application.config.config import Config
from petisco.application.singleton import Singleton
from petisco.logger.not_implemented_logger import NotImplementedLogger

DEPLOY_TOPIC = "deploy"


@dataclass
class Petisco(metaclass=Singleton):
    app_name: str
    app_version: str
    logger: ILogger = NotImplementedLogger()
    application: IApplication = None
    services_provider: Callable = None
    repositories_provider: Callable = None
    event_manager_provider: Callable = None
    options: Dict[str, Any] = None
    info: Dict = None
    _persistence_models: Dict[str, Any] = False
    persistence_configured: bool = False
    config: Config = None
    event_topic: str = None

    def __init__(self, config: Config):
        self.config = config
        self.app_name = config.app_name
        self.app_version = config.app_version
        self.logger = config.get_logger()
        self.info = {}
        self.set_persistence(config)
        self.set_infrastructure(config)
        self.options = config.options

        if self.info:
            self.logger.log(INFO, f"Info: {self.info}")

        if self.options:
            self.logger.log(INFO, f"Options: {self.options}")

        if self.config.config_infrastructure.publish_deploy_event_func:
            event_manager = self.config.config_infrastructure.event_manager_provider()
            event = ServiceDeployed(
                app_name=self.app_name, app_version=self.app_version
            )
            event_manager.publish(topic=DEPLOY_TOPIC, event=event)

    @staticmethod
    def get_instance():
        try:
            return Petisco()
        except Exception as e:  # noqa E722
            frame_info = inspect.stack()[1]
            raise ImportError(
                f"Petisco must be configured when the application starts. \n"
                f"Petisco has been requested before its initial configuration.\n "
                f"Following code must be executed after Petisco initialization:\n"
                f"\tfilename: {frame_info.filename}\n"
                f"\tlineno: {frame_info.lineno}\n"
                f"\tfunction: {frame_info.function}\n"
                f"\tcode_context: {frame_info.code_context}"
            )

    @staticmethod
    def from_filename(filename: str):
        """
       Parameters
       ----------
       filename
           YAML-based configuration file (default petisco.yml)
        """
        print(f"Loading petisco from: {filename}")

        config = Config.from_filename(filename).unwrap_or_throw()

        return Petisco(config=config)

    def set_persistence(self, config):
        config_persistence = config.config_persistence

        if config_persistence.config:
            config_persistence.config()
            self.persistence_configured = True

        self._persistence_models = config_persistence.get_models()

    def set_infrastructure(self, config):
        config_infrastructure = config.config_infrastructure
        if not config_infrastructure:
            return
        if config_infrastructure.config_dependencies:
            config_infrastructure.config_dependencies()

        if config_infrastructure.services_provider:
            self.services_provider = config_infrastructure.services_provider

            info_services = {}
            for key, service in self.services_provider().items():
                if hasattr(service, "info"):
                    info_services[key] = service.info()
                else:
                    raise TypeError(
                        f"Service with key {key} ({type(service)}) must implement info"
                    )
            self.info["services"] = info_services
        if config_infrastructure.repositories_provider:
            self.repositories_provider = config_infrastructure.repositories_provider

            info_repositories = {}
            for key, repository in self.repositories_provider().items():
                if hasattr(repository, "info"):
                    info_repositories[key] = repository.info()
                else:
                    raise TypeError(
                        f"Repository with key {key} ({type(repository)}) must implement info"
                    )

            self.info["repositories"] = info_repositories
        if config_infrastructure.event_manager_provider:
            self.event_manager_provider = config_infrastructure.event_manager_provider
            event_manager = self.event_manager_provider()
            if hasattr(event_manager, "info"):
                self.info["event_managers"] = event_manager.info()
            else:
                raise TypeError(
                    f"Given event_manager ({type(event_manager)}) must implement info"
                )

        if config_infrastructure.event_topic:
            self.event_topic = config_infrastructure.event_topic

    def start(self):
        self.config.get_application().start()

    def get_app(self):
        return self.config.get_application().get_app()

    @staticmethod
    def services() -> DotMap:
        return DotMap(Petisco.get_instance().services_provider())

    @staticmethod
    def repositories() -> DotMap:
        return DotMap(Petisco.get_instance().repositories_provider())

    @staticmethod
    def persistence_models() -> DotMap:
        persistence_models = {}
        try:
            persistence_models = Petisco.get_instance()._persistence_models
        except:  # noqa E722
            pass
        return persistence_models

    @staticmethod
    def persistence_session_scope():
        from petisco.persistence.sqlalchemy.sqlalchemy_session_scope import (
            session_scope,
        )

        return session_scope

    @staticmethod
    def event_manager() -> IEventManager:
        return Petisco.get_instance().event_manager_provider()

    @staticmethod
    def providers():
        return Petisco.services(), Petisco.repositories(), Petisco.event_manager()



from typing import Any

from dataclasses import dataclass
from dataclasses_json import dataclass_json

from petisco.domain.aggregate_roots.info_id import InfoId


@dataclass_json
@dataclass
class LogMessage:
    layer: str = None
    operation: str = None
    info_id: InfoId = None
    message: Any = None



from petisco.events.event import Event

EVENT_SERVICE_DEPLOYED_VERSION = "1.0.0"


class ServiceDeployed(Event):
    app_name: str = None
    app_version: str = None

    def __init__(
        self,
        app_name: str,
        app_version: str,
        event_version: str = EVENT_SERVICE_DEPLOYED_VERSION,
    ):
        self.app_name = app_name
        self.app_version = app_version
        self.event_version = event_version

        super().__init__()



import importlib
from typing import Callable


def get_function_from_string(function_string: str) -> Callable:
    if not function_string:
        return None
    mod_name, func_name = function_string.rsplit(".", 1)
    mod = importlib.import_module(mod_name)
    func = getattr(mod, func_name)
    return func



from typing import Optional

from dataclasses import dataclass
from dataclasses_json import dataclass_json

from petisco.application.config.get_funtion_from_string import get_function_from_string


@dataclass_json
@dataclass
class ConfigInfrastructure:
    config_dependencies_func: Optional[str] = None
    services_provider_func: Optional[str] = None
    repositories_provider_func: Optional[str] = None
    event_manager_provider_func: Optional[str] = None
    publish_deploy_event_func: Optional[bool] = False
    event_topic: Optional[str] = None

    @property
    def config_dependencies(self):
        return get_function_from_string(self.config_dependencies_func)

    @property
    def config_persistence(self):
        return get_function_from_string(self.config_persistence_func)

    @property
    def services_provider(self):
        return get_function_from_string(self.services_provider_func)

    @property
    def repositories_provider(self):
        return get_function_from_string(self.repositories_provider_func)

    @property
    def event_manager_provider(self):
        return get_function_from_string(self.event_manager_provider_func)



from typing import Callable, Optional

from dataclasses import dataclass
from dataclasses_json import dataclass_json

from petisco.application.config.get_funtion_from_string import get_function_from_string


@dataclass_json
@dataclass
class ConfigLogger:
    selected_logger: Optional[str] = "not_implemented_logger"
    name: Optional[str] = None
    format: Optional[str] = None
    config_func: Optional[str] = None

    def get_config_func(self) -> Callable:
        return get_function_from_string(self.config_func)



import importlib
from typing import Optional, Dict, Any

from dataclasses import dataclass
from dataclasses_json import dataclass_json

from petisco.application.config.get_funtion_from_string import get_function_from_string


@dataclass_json
@dataclass
class ConfigPersistence:
    config_func: Optional[str] = None
    models: Optional[Dict[str, Any]] = None

    @property
    def config(self):
        return get_function_from_string(self.config_func)

    def get_models(self):
        loaded_models = {}
        for name, model_string in self.models.items():
            mod_name, model_name = model_string.rsplit(".", 1)
            mod = importlib.import_module(mod_name)
            loaded_models[name] = getattr(mod, model_name)
        return loaded_models
